from libcpp.vector cimport vector

from CyRK.cy.common cimport OptimizeInfo, CyrkErrorCodes
from CyRK.cy.cysolver_api cimport CySolverDense
from CyRK.cy.events cimport EventFunc, EventFuncWithInst, Event

cdef extern from "c_common.cpp" nogil:
    pass

cdef extern from "c_events.cpp" nogil:
    pass

cdef extern from "dense.cpp" nogil:
    pass

cdef extern from "c_brentq.cpp" nogil:

    cdef double c_brentq(
        EventFuncWithInst func,
        double xa,
        double xb,
        double xtol,
        double rtol,
        size_t iter,
        vector[char]& func_data_vec,
        OptimizeInfo* solver_stats,
        Event* event_ptr,
        CySolverDense* dense_func)


from CyRK.cy.common cimport DiffeqFuncType, PreEvalFunc, MAX_STEP, CyrkErrorCodes, round_to_2, OptimizeInfo, MAX_SIZET_SIZE
from CyRK.optimize.brentq cimport c_brentq
from CyRK.cy.events cimport Event, EventFunc
from CyRK.cy.cysolver_api cimport cysolve_ivp, cysolve_ivp_gil, cysolve_ivp_noreturn, CySolverResult, WrapCySolverResult, CySolverBase, CySolveOutput, ODEMethod
from CyRK.cy.pysolver cimport PySolver
from CyRK.cy.helpers cimport interpolate_from_solution_list


import numpy as np
from CyRK import pysolve_ivp
import gc


def diffeq(dy, t, y):
    dy[0] = (1. - 0.01 * y[1]) * y[0]
    dy[1] = (0.02 * y[0] - 1.) * y[1]

initial_conds = np.asarray((20., 20.), dtype=np.float64, order='C')
time_span = (0., 10.)
time_span_large = (0., 1000.)
rtol = 1.0e-4
atol = 1.0e-7

def run_diffeq_array_return():
    result = pysolve_ivp(diffeq, time_span, initial_conds, method="RK45", rtol=rtol, atol=atol, pass_dy_as_arg=True)

    t = result.t
    y = result.y
    del result
    gc.collect()
    return t, y

def run_diffeq_result_return():
    result = pysolve_ivp(diffeq, time_span, initial_conds, method="RK45", rtol=rtol, atol=atol, pass_dy_as_arg=True)
    return result

def test_pysolve_array_lifetimes():
    """ Tests that the underlying arrays returned by the WrapCySolverResult class can outlive the class instance. """
    
    # Run in a loop because sometimes memory corruptions can take a while to appear
    for i in range(50):
        # Get result from the result instance class
        result = run_diffeq_result_return()

        # Get arrays from the array return
        t, y = run_diffeq_array_return()

        # Check that they contain the same data
        assert result.y.size == y.size
        assert result.t.size == t.size

        assert np.allclose(result.y, y)
        assert np.allclose(result.t, t)



from libcpp cimport bool as cpp_bool
cimport cpython.ref as cpy_ref

from CyRK.utils.memory cimport unique_ptr, make_unique
from CyRK.cy.pysolver_cyhook cimport DiffeqMethod, PyEventMethod
from CyRK.cy.cysolver_api cimport ODEMethod, CySolverResult, NowStatePointers, WrapCySolverResult

cimport numpy as cnp
cnp.import_array()

# =====================================================================================================================
# Import the C++ cysolve_ivp helper function
# =====================================================================================================================
cdef class PySolver(WrapCySolverResult):

    cdef object diffeq_func
    cdef tuple args
    cdef cpp_bool use_args
    cdef cpp_bool pass_dy_as_arg

    cdef size_t num_y
    cdef size_t num_dy

    cdef cnp.ndarray y_tmp_arr
    cdef double[::1] y_tmp_view
    cdef cnp.ndarray y_now_arr
    cdef double[::1] y_now_view
    cdef cnp.ndarray dy_now_arr
    cdef double[::1] dy_now_view

    # State attributes
    cdef double* y_now_ptr
    cdef double* t_now_ptr
    cdef double* dy_now_ptr

    # Event data
    cdef list events_list

    cdef void set_state(self, NowStatePointers* solver_state_ptr) noexcept
    cpdef set_pydiffeq(
        self,
        object diffeq_func,
        tuple args,
        size_t num_y,
        size_t num_dy,
        bint pass_dy_as_arg = *
        )
    cpdef set_problem_parameters(
        self,
        object py_diffeq,
        tuple time_span,
        const double[::1] y0,
        str method = *,
        const double[::1] t_eval = *,
        bint dense_output = *,
        object events = *,
        tuple args = *,
        size_t expected_size = *,
        size_t num_extra = *,
        double first_step = *,
        double max_step = *,
        rtol = *,
        atol = *,
        size_t max_num_steps = *,
        size_t max_ram_MB = *,
        bint pass_dy_as_arg = *
        )
    cdef void diffeq(self) noexcept
    cdef double check_pyevent(
        self,
        size_t event_index,
        double t, 
        double* y_ptr) noexcept



# distutils: language = c++
# cython: boundscheck=False, wraparound=False, nonecheck=False, cdivision=True, initializedcheck=False
from libc.string cimport memcpy
from libcpp.cmath cimport fmin, fabs
from libcpp.vector cimport vector

from CyRK.cy.common cimport INF, EPS_100, CyrkErrorCodes, CyrkErrorMessages, find_expected_size, dbl_NAN, MAX_SIZET_SIZE
from CyRK.cy.cysolver_api cimport ProblemConfig, RKConfig, CySolverBase
from CyRK.cy.events cimport Event, EventFunc

import numpy as np
cimport numpy as cnp
cnp.import_array()

# =====================================================================================================================
# PySolver Class (holds the intergrator class and reference to the python diffeq function)
# =====================================================================================================================
cdef class PySolver(WrapCySolverResult):

    cpdef set_pydiffeq(
            self,
            object diffeq_func,
            tuple args,
            size_t num_y,
            size_t num_dy,
            bint pass_dy_as_arg = False
            ):
        
        # Install differential equation function and any additional args
        self.diffeq_func = diffeq_func
        if args is None:
            self.args     = None
            self.use_args = False
        else:
            if len(args) == 0:
                # Empty tuple provided. Don't use args.
                self.args     = None
                self.use_args = False
            else:
                self.args     = args
                self.use_args = True
        
        # Build python-safe arrays
        self.num_y  = num_y
        self.num_dy = num_dy

        if pass_dy_as_arg:
            self.pass_dy_as_arg = True
        else:
            self.pass_dy_as_arg = False

    cpdef set_problem_parameters(
            self,
            object py_diffeq,
            tuple time_span,
            const double[::1] y0,
            str method = 'RK45',
            const double[::1] t_eval = None,
            bint dense_output = False,
            object events = None,
            tuple args = None,
            size_t expected_size = 0,
            size_t num_extra = 0,
            double first_step = 0.0,
            double max_step = INF,
            rtol = 1.0e-3,
            atol = 1.0e-6,
            size_t max_num_steps = 0,
            size_t max_ram_MB = 2000,
            bint pass_dy_as_arg = False
            ):
        # Parse method
        method = method.lower()
        cdef ODEMethod integration_method = ODEMethod.RK45
        if method == "rk23":
            integration_method = ODEMethod.RK23
        elif method == "rk45":
            integration_method = ODEMethod.RK45
        elif method == 'dop853':
            integration_method = ODEMethod.DOP853
        else:
            raise NotImplementedError(
                "ERROR: `PySolver::set_problem_parameters` - "
                f"Unknown or unsupported integration method provided: {method}.\n"
                f"Supported methods are: RK23, RK45, DOP853."
                )

        cdef CySolverResult* cyresult_ptr = self.cyresult_uptr.get()
        if self.cyresult_uptr:
            # Already have a solution class created.
            # Is this the same integration method? If it is then we can reuse the solution class.
            if integration_method != cyresult_ptr.integrator_method:
                # Need to make a new cysolver object
                self.build_cyresult(integration_method)
        else:
            self.build_cyresult(integration_method)
        cyresult_ptr = self.cyresult_uptr.get()
        cdef ProblemConfig* base_config_ptr = cyresult_ptr.config_uptr.get()
        
        if not cyresult_ptr:
            raise RuntimeError("ERROR: `PySolver::set_problem_parameters` - CySolverResult was not constructed.")
        cdef CySolverBase* cysolver_ptr = cyresult_ptr.solver_uptr.get()
        if not cysolver_ptr:
            raise AttributeError("ERROR: `PySolver::set_problem_parameters` - CySolver not set within CySolverResult object.")

        # Parse y0
        cdef size_t i, y_i
        cdef size_t num_y  = y0.size
        cdef size_t num_dy = num_y + num_extra  
        cdef vector[double] y0_vec = vector[double](num_y)
        for i in range(num_y):
            y0_vec[i] = y0[i]
        
        # We need to set the number of ys now because we need the now state pointers. 
        # These pointers could change which memory they are pointing to if by setting the num_y later 
        # causes a realloc of the underlying vectors.
        cdef CyrkErrorCodes setup_status = cysolver_ptr.resize_num_y(num_y, num_dy)
        if setup_status != CyrkErrorCodes.NO_ERROR:
            raise Exception("ERROR: `PySolver::set_problem_parameters` - Error raised while setting number of ys: {setup_status}.")

        # Setup python diffeq
        self.set_pydiffeq(py_diffeq, args, num_y, num_dy, pass_dy_as_arg)

        # Pass python pointers to C++ classes.
        cdef DiffeqMethod diffeq_func = <DiffeqMethod>self.diffeq
        cysolver_ptr.set_cython_extension_instance(<cpy_ref.PyObject*>self, diffeq_func)

        # Pull in current state pointers from CySolver C++ object to this object so that `self.diffeq` can update the "now" attributes.
        cdef NowStatePointers solver_state = cysolver_ptr.get_now_state()
        self.set_state(&solver_state)

        # Update other configurations now
        if len(time_span) != 2:
            raise AttributeError("ERROR: `PySolver::set_problem_parameters` - Unexpected size found for the provided `time_span`.")
        cdef double t_start = time_span[0]
        cdef double t_end   = time_span[1]

        # Update configurations
        cdef RKConfig* problem_config_ptr = <RKConfig*>base_config_ptr

        # Pass python pointers to C++ classes.
        problem_config_ptr.cython_extension_class_instance = <cpy_ref.PyObject*>self
        problem_config_ptr.py_diffeq_method                = <DiffeqMethod>self.diffeq
        
        # Set required arguments.
        # diffeq_ptr - unused for PySolver.
        problem_config_ptr.t_start = t_start
        problem_config_ptr.t_end = t_end
        problem_config_ptr.y0_vec = y0_vec

        # Parse Events
        problem_config_ptr.check_events = False
        problem_config_ptr.events_vec.resize(0)
        cdef size_t num_events = 0
        cdef int event_direction 
        cdef size_t event_max_allowed
        cdef object terminal
        cdef EventFunc cyevent_func = NULL
        cdef CyrkErrorCodes event_setup_status = CyrkErrorCodes.NO_ERROR
        cdef PyEventMethod pyevent_check_func = <PyEventMethod>self.check_pyevent
        cdef object py_event
        if events is not None:
            self.events_list = list()
            if type(events) in (tuple, list):
                for py_event in events:
                    self.events_list.append(py_event)
                    num_events += 1
            else:
                self.events_list.append(events)
                num_events = 1

            # Build events vector
            for i in range(num_events):
                # Pull out properties the user may have set.
                event_direction = getattr(self.events_list[i], 'direction', 0)
                terminal        = getattr(self.events_list[i], 'terminal', None)
                if terminal is None:
                    event_max_allowed = getattr(self.events_list[i], 'max_allowed', MAX_SIZET_SIZE)
                else:
                    event_max_allowed = <size_t>terminal
            
                # Build events with null pointers - we won't use the cython version of the event checker function
                problem_config_ptr.events_vec.emplace_back(cyevent_func, event_max_allowed, event_direction)
                
                # Set up other properties so it can work with PySolver
                problem_config_ptr.events_vec[i].pyevent_index = i
                event_setup_status = problem_config_ptr.events_vec[i].set_cython_extension_instance(
                    <cpy_ref.PyObject*>self,
                    pyevent_check_func
                    )
                if event_setup_status != CyrkErrorCodes.NO_ERROR:
                    raise Exception(f"ERROR: `PySolver.set_problem_parameters` - Failed to setup pyhooks in Event class (event index {i}).")
            
            # Setup a python-safe array used in pyevent solver.
            self.y_tmp_arr  = np.empty(self.num_dy, dtype=np.float64, order='C')
            self.y_tmp_view = self.y_tmp_arr

        # Parse t_eval
        cdef size_t t_eval_size = 0
        if t_eval is None:
            problem_config_ptr.t_eval_provided = False
            problem_config_ptr.t_eval_vec.resize(0)
        else:
            t_eval_size = t_eval.size
            problem_config_ptr.t_eval_provided = True
            problem_config_ptr.t_eval_vec.resize(t_eval_size)
            for i in range(t_eval_size):
                problem_config_ptr.t_eval_vec[i] = t_eval[i]
        
        # Parse rtol
        cdef size_t rtol_size = 0
        if type(rtol) == float:
            problem_config_ptr.rtols.resize(1)
            problem_config_ptr.rtols[0] = rtol
        else:
            rtol_size = rtol.size
            if rtol_size > 1 and rtol_size != num_y:
                raise AttributeError("ERROR: `PySolver.set_problem_parameters` - Provided rtol array size must be 1 or the number of dependent variables.")
            
            problem_config_ptr.rtols.resize(rtol_size)
            for y_i in range(rtol_size):
                problem_config_ptr.rtols[y_i] = rtol[y_i]
        
        # Parse atol
        cdef size_t atol_size = 0
        if type(atol) == float:
            problem_config_ptr.atols.resize(1)
            problem_config_ptr.atols[0] = atol
        else:
            atol_size = atol.size
            if atol_size > 1 and atol_size != num_y:
                raise AttributeError("ERROR: `PySolver.set_problem_parameters` - Provided atol array size must be 1 or the number of dependent variables.")

            problem_config_ptr.atols.resize(atol_size)
            for y_i in range(atol_size):
                problem_config_ptr.atols[y_i] = atol[y_i]
        
        # Parse expected size
        cdef size_t expected_size_touse = expected_size
        cdef double rtol_tmp
        cdef double min_rtol = INF
        if expected_size_touse == 0:
            for i in range(problem_config_ptr.rtols.size()):
                rtol_tmp = problem_config_ptr.rtols[i]
                if rtol_tmp < EPS_100:
                    rtol_tmp = EPS_100
                min_rtol = fmin(min_rtol, rtol_tmp)
            expected_size_touse = find_expected_size(num_y, num_extra, fabs(t_end - t_start), min_rtol)
        problem_config_ptr.expected_size = expected_size_touse

        # Parse first step size
        if first_step < 0.0:
            raise AttributeError("ERROR: `PySolver::set_problem_parameters` - First step size must be a postive float (or 0.0 to use automatic finder).")
        problem_config_ptr.first_step_size = first_step
    
        # Parse maximum step size
        if max_step <= 0.0:
            raise AttributeError("ERROR: `PySolver::set_problem_parameters` - Maximum step size must be a postive float.")
        problem_config_ptr.max_step_size = max_step

        # Parse other flags
        problem_config_ptr.capture_dense_output = dense_output
        problem_config_ptr.force_retain_solver  = True # For now we are going to keep solvers it is a small memory hit but makes everything much easier to debug and avoids crashes.
        problem_config_ptr.num_extra            = num_extra
        problem_config_ptr.capture_extra        = num_extra > 0
        problem_config_ptr.max_num_steps        = max_num_steps
        problem_config_ptr.max_ram_MB           = max_ram_MB

        # Load config into cysolution
        cdef CyrkErrorCodes status_code = cyresult_ptr.setup()

        if status_code != CyrkErrorCodes.NO_ERROR:
            raise Exception(
                f"ERROR: `PySolver::set_problem_parameters` - Error during config setup. Error Code: {status_code}. "
                f"Message: {CyrkErrorMessages.at(status_code).decode('utf-8')}")
    
    cdef void set_state(self, NowStatePointers* solver_state_ptr) noexcept:
        
        self.t_now_ptr  = solver_state_ptr.t_now_ptr
        self.y_now_ptr  = solver_state_ptr.y_now_ptr
        self.dy_now_ptr = solver_state_ptr.dy_now_ptr

        # Create memoryviews of the pointers
        self.y_now_view  = <double[:self.num_y]>self.y_now_ptr

        # Create numpy arrays which will be passed to the python diffeq.
        # We need to make sure that this is a not a new ndarray, but one that points to the same data. 
        # That is why we use `PyArray_SimpleNewFromData` instead of a more simple `asarray`.
        # Note that it is not safe to return these arrays outside of this class because they may get deallocated while
        # the numpy array still points to the underlying memory.
        cdef cnp.npy_intp[1] shape
        cdef cnp.npy_intp* shape_ptr = &shape[0]
        shape_ptr[0] = <cnp.npy_intp>self.num_y
        
        self.y_now_arr = cnp.PyArray_SimpleNewFromData(1, shape_ptr, cnp.NPY_DOUBLE, self.y_now_ptr)
        
        # Do the same for dy if the user provided the appropriate kind of differential equation.
        if self.pass_dy_as_arg:
            self.dy_now_view = <double[:self.num_dy]>self.dy_now_ptr
            shape[0]         = <cnp.npy_intp>self.num_dy  # dy may have a larger shape than y
            self.dy_now_arr  = cnp.PyArray_SimpleNewFromData(1, shape_ptr, cnp.NPY_DOUBLE, self.dy_now_ptr)   

    cdef void diffeq(self) noexcept:
        # Run python diffeq
        if self.pass_dy_as_arg:
            if self.use_args:
                self.diffeq_func(self.dy_now_arr, self.t_now_ptr[0], self.y_now_arr, *self.args)
            else:
                self.diffeq_func(self.dy_now_arr, self.t_now_ptr[0], self.y_now_arr)
        else:
            if self.use_args:
                self.dy_now_view = self.diffeq_func(self.t_now_ptr[0], self.y_now_arr, *self.args)
            else:
                self.dy_now_view = self.diffeq_func(self.t_now_ptr[0], self.y_now_arr)
            # Since we do not have a static dy that we can pass to the function and use in the solver we must copy over
            # the values from the newly created dy memory view
            # Note that num_dy may be larger than num_y if the user is capturing extra output during integration.
            memcpy(self.dy_now_ptr, &self.dy_now_view[0], sizeof(double) * self.num_dy)
    
    cdef double check_pyevent(
            self,
            size_t event_index,
            double t, 
            double* y_ptr) noexcept:

        # Convert pointer to python object so it can be passed to the python event function
        # TODO: Explore ways to improve performance of this function.
        cdef size_t y_i
        for y_i in range(self.num_dy):
            self.y_tmp_view[y_i] = y_ptr[y_i]
        
        cdef double result = dbl_NAN
        if self.use_args:
            result = self.events_list[event_index](t, self.y_tmp_arr, *self.args)
        else:
            result = self.events_list[event_index](t, self.y_tmp_arr)
        return result

# =====================================================================================================================
# PySolver wrapper function
# =====================================================================================================================
def pysolve_ivp(
        object py_diffeq,
        tuple time_span,
        const double[::1] y0,
        str method = 'RK45',
        const double[::1] t_eval = None,
        bint dense_output = False,
        object events = None,
        tuple args = None,
        size_t expected_size = 0,
        size_t num_extra = 0,
        double first_step = 0.0,
        double max_step = INF,
        rtol = 1.0e-3,
        atol = 1.0e-6,
        size_t max_num_steps = 0,
        size_t max_ram_MB = 2000,
        bint pass_dy_as_arg = False,
        PySolver solution_reuse = None
        ):

    # Build PySolver solution storage.
    # These cython extension classes are created as python objects so they have reference counting
    # so it is safe to return objects created in this function without worry of memory leaks.
    if solution_reuse is None:
        solution_reuse = PySolver()

    # Load in user-provided parameters
    solution_reuse.set_problem_parameters(
            py_diffeq,
            time_span,
            y0,
            method,
            t_eval,
            dense_output,
            events,
            args,
            expected_size,
            num_extra,
            first_step,
            max_step,
            rtol,
            atol,
            max_num_steps,
            max_ram_MB,
            pass_dy_as_arg)
    
    ##
    # Run the integrator!
    ##
    solution_reuse.solve()
    
    # Return the results.
    return solution_reuse



# distutils: language = c++
# cython: boundscheck=False, wraparound=False, nonecheck=False, cdivision=True, initializedcheck=False
from libc.math cimport floor
from libcpp.utility cimport move

import numpy as np
cimport numpy as cnp
cnp.import_array()

# =====================================================================================================================
# Import CySolverResult (container for integration results)
# =====================================================================================================================
cdef class WrapCySolverResult:

    def __dealloc__(self):
        self.cyresult_uptr.reset()

    cdef void build_cyresult(
            self,
            ODEMethod integrator_method):
        
        # Delete any CySolver Result object that may have already been set. 
        if self.cyresult_uptr:
            self.cyresult_uptr.reset()

        self.set_cyresult_pointer(make_unique[CySolverResult](integrator_method))

    cdef void set_cyresult_pointer(self, unique_ptr[CySolverResult] cyresult_uptr_):
        if not cyresult_uptr_:
            raise AttributeError("ERROR: `WrapCySolverResult::Constructor` - Provided CySolverResult is Null.")

        # Store c++ based result and pull out key information
        self.cyresult_uptr = move(cyresult_uptr_)
        cdef CySolverResult* cyresult_ptr = self.cyresult_uptr.get()

        self.finalize()
    
    cdef set_problem_config(self, ProblemConfig* new_problem_config_ptr):
        if not self.cyresult_uptr:
            raise AttributeError("ERROR: `WrapCySolverResult::set_problem_config` - CySolverResult not set.")
        self.cyresult_uptr.get().setup(new_problem_config_ptr)
    
    cpdef solve(self):
        if not self.cyresult_uptr:
            raise AttributeError("ERROR: `WrapCySolverResult::solve` - CySolverResult not set.")
        cdef CySolverResult* cyresult_ptr = self.cyresult_uptr.get()
        cyresult_ptr.solve()

        self.finalize()
    
    cpdef finalize(self):
        # Convert solution to pointers and views
        cdef CySolverResult* cyresult_ptr = self.cyresult_uptr.get()

        cdef double[::1] dbl_view
        cdef double* dbl_ptr
        cdef size_t i, size1, size2

        if cyresult_ptr:
            if cyresult_ptr.size > 0:
                self.time_ptr  = &cyresult_ptr.time_domain_vec[0]
                self.y_ptr     = &cyresult_ptr.solution[0]
                self.time_view = <double[:self.size]>self.time_ptr
                self.y_view    = <double[:self.size * self.num_dy]>self.y_ptr
            
            # Handle event arrays.
            # Event times and data are stored in a vector of vectors (vector[vector[double]])
            # Data is of size dy since extra output is also grabbed.
            if cyresult_ptr.num_events > 0:
                # Event arrays will have a different size for each event since one event may trigger more than another.
                # We will store each one d array in a numpy array that is then stored in a python list; one array for each event.
                
                # Initialize lists
                self.list_t_events = list()
                self.list_y_events = list()

                # Step through each event building our numpy arrays.
                for i in range(cyresult_ptr.num_events):
                    size1 = cyresult_ptr.event_times[i].size()
                    if size1 > 0:
                        dbl_ptr = &cyresult_ptr.event_times[i][0]
                        dbl_view = <double[:size1]>dbl_ptr
                        # Convert to numpy array and store in our list.
                        self.list_t_events.append(np.copy(np.asarray(dbl_view, dtype=np.float64, order='C')))

                        # Repeat for y data but now we need to reshape.
                        size2 = size1 * self.num_dy
                        dbl_ptr = &cyresult_ptr.event_states[i][0]
                        dbl_view = <double[:size2]>dbl_ptr
                        self.list_y_events.append(np.copy(np.asarray(dbl_view, dtype=np.float64, order='C')).reshape((size1, self.num_dy)).T)
                    else:
                        # No triggered events. Return empty arrays
                        self.list_t_events.append(np.array([], dtype=np.float64, order='C'))
                        self.list_y_events.append(np.array([], dtype=np.float64, order='C'))

    def call(self, double t):
        """ Call the dense output interpolater and return y """

        if not self.cyresult_uptr:
            raise AttributeError("ERROR: `WrapCySolverResult::call` - CySolverResult is Null.")
        cdef CySolverResult* cyresult_ptr = self.cyresult_uptr.get()
        if not cyresult_ptr.capture_dense_output:
            raise AttributeError("ERROR: `WrapCySolverResult::call` - Can not make a call to WrapCySolverResult when dense_output set to False.")

        y_interp_array = np.empty(cyresult_ptr.num_dy, dtype=np.float64, order='C')
        cdef double[::1] y_interp_view = y_interp_array
        cdef double* y_interp_ptr      = &y_interp_view[0]

        cyresult_ptr.call(t, y_interp_ptr)
        
        return y_interp_array.reshape(self.num_dy, 1)
    
    def call_vectorize(self, double[::1] t_view):
        """ Call the dense output interpolater and return y """
        
        if not self.cyresult_uptr:
            raise AttributeError("ERROR: `WrapCySolverResult::call_vectorize` - CySolverResult is Null.")
        cdef CySolverResult* cyresult_ptr = self.cyresult_uptr.get()
        if not cyresult_ptr.capture_dense_output:
            raise AttributeError("ERROR: `WrapCySolverResult::call_vectorize` - Can not call WrapCySolverResult when dense_output set to False.")

        cdef size_t len_t = len(t_view)

        y_interp_array = np.empty(cyresult_ptr.num_dy * len_t, dtype=np.float64, order='C')
        cdef double[::1] y_interp_view = y_interp_array
        cdef double* y_interp_ptr      = &y_interp_view[0]
        cdef double* t_array_ptr       = &t_view[0]

        cyresult_ptr.call_vectorize(t_array_ptr, len_t, y_interp_ptr)
        return y_interp_array.reshape(len_t, cyresult_ptr.num_dy).T

    def print_diagnostics(self):
        
        cdef size_t i
        
        if not self.cyresult_uptr:
            raise AttributeError("ERROR: `WrapCySolverResult::print_diagnostics` - CySolverResult is Null.")
        cdef CySolverResult* cyresult_ptr = self.cyresult_uptr.get()
        cdef ProblemConfig* config_ptr = cyresult_ptr.config_uptr.get()
    
        cdef str diagnostic_str = ''
        from CyRK import __version__
        cdef str direction_str = 'Forward'
        if cyresult_ptr.direction_flag == 0:
            direction_str = 'Backward'
        
        diagnostic_str += f'----------------------------------------------------\n'
        diagnostic_str += f'CyRK v{__version__} - WrapCySolverResult Diagnostic.\n'
        diagnostic_str += f'----------------------------------------------------\n'
        diagnostic_str += f'# of y:       {self.num_y}.\n'
        diagnostic_str += f'# of dy:      {self.num_dy}.\n'
        diagnostic_str += f'# of events:  {self.num_events}.\n'
        diagnostic_str += f'Success:      {self.success}.\n'
        diagnostic_str += f'Error Code:   {self.error_code}.\n'
        diagnostic_str += f'Size:         {self.size}.\n'
        diagnostic_str += f'Steps Taken:  {self.steps_taken}.\n'
        diagnostic_str += f'Event Termination:  {self.event_terminated}.\n'
        if self.event_terminated:
            diagnostic_str += f'Event Term. Index:  {self.event_terminate_index}.\n'
        diagnostic_str += f'Integrator Message:\n\t{self.message}\n'
        diagnostic_str += f'\n----------------- CySolverResult -------------------\n'
        diagnostic_str += f'Capture Extra:          {cyresult_ptr.capture_extra}.\n'
        diagnostic_str += f'Capture Dense Output:   {cyresult_ptr.capture_dense_output}.\n'
        diagnostic_str += f'Integration Direction:  {direction_str}.\n'
        diagnostic_str += f'Integration Method:     {self.integration_method}.\n'
        diagnostic_str += f'# of Interpolates:      {cyresult_ptr.num_interpolates}.\n'

        if cyresult_ptr.num_events > 0:
            diagnostic_str += f'\n------------------- Event Data ---------------------\n'
            for i in range(cyresult_ptr.num_events):
                diagnostic_str += f'Event {i}:\n'
                diagnostic_str += f'\tDirection:      {config_ptr.events_vec[i].direction}.\n'
                diagnostic_str += f'\tMax Allowed:    {config_ptr.events_vec[i].max_allowed}.\n'
                diagnostic_str += f'\tNum Triggers:   {config_ptr.events_vec[i].current_count}.\n'
                diagnostic_str += f'\tLast Trigger t: {config_ptr.events_vec[i].last_root:0.4e}.\n'

        cdef CySolverBase* cysolver = cyresult_ptr.solver_uptr.get()
        cdef size_t num_y
        cdef size_t num_dy
        cdef size_t dbl_i
        cdef size_t args_size
        cdef size_t args_size_dbls
        cdef double* args_dbl_ptr = NULL
        cdef char* args_char_ptr = NULL
        if cysolver:
            num_y  = cysolver.num_y
            num_dy = cysolver.num_dy
            diagnostic_str += f'\n------------------ CySolverBase --------------------\n'
            diagnostic_str += f'Status #:   {cyresult_ptr.status}.\n'
            diagnostic_str += f'Status:     {CyrkErrorMessages.at(cyresult_ptr.status).decode("utf-8")}.\n'
            diagnostic_str += f'# of y:     {num_y}.\n'
            diagnostic_str += f'# of dy:    {num_dy}.\n'
            diagnostic_str += f'PySolver:   {cysolver.use_pysolver}.\n'
            diagnostic_str += f'---- Current State Info ----\n'
            diagnostic_str += f't_now:      {cysolver.t_now}.\n'
            diagnostic_str += f'y_now:\n'
            for i in range(num_y):
                diagnostic_str += f'\ty{i}  = {cysolver.y_now_ptr[i]:0.5e}.\n'
            diagnostic_str += f'dy_now:\n'
            for i in range(num_dy):
                diagnostic_str += f'\tdy{i} = {cysolver.dy_now_ptr[i]:0.5e}.\n'
            diagnostic_str += f'End of Current State Info.\n'
            # We want a way to print the value of the args but we have no idea its structure or types
            # of its contents. For now just assume they are all doubles so we can display something.
            # Also display the value of the raw chars.
            diagnostic_str += f'---- Additional Argument Info ----\n'
            args_size      = config_ptr.args_vec.size()
            args_char_ptr  = config_ptr.args_vec.data()
            args_dbl_ptr   = <double*>args_char_ptr
            args_size_dbls = <size_t>floor(args_size / sizeof(double))
            diagnostic_str += f'args size (bytes):   {args_size}.\n'
            diagnostic_str += f'args size (doubles): {args_size_dbls}.\n'
            if not args_char_ptr:
                diagnostic_str += 'Args Pointer is Null.\n'
            else:
                dbl_i = 0
                if args_size > 0:
                    for i in range(args_size):
                        if i % 8 == 0:
                            # New section of 8 bytes.
                            diagnostic_str += f'\n{hex(args_char_ptr[i])}'
                        elif i % 8 == 3:
                            diagnostic_str += f' {hex(args_char_ptr[i])}\n'
                        elif i % 8 == 7:
                            diagnostic_str += f' {hex(args_char_ptr[i])}\n'
                            diagnostic_str += f'As Double: {args_dbl_ptr[dbl_i]:0.5e}\n'
                            dbl_i += 1
                        else:
                            diagnostic_str += f' {hex(args_char_ptr[i])}'
            diagnostic_str += f'End of Additional Argument Info.\n'
        else:
            diagnostic_str += f'CySolver instance was deleted or voided.\n'

        diagnostic_str += f'\n-------------- Diagnostic Complete -----------------\n'
        print(diagnostic_str)

    def __dealloc__(self):
        # Deallocate any heap allocated memory
        if self.cyresult_uptr:
            self.cyresult_uptr.reset()

    @property
    def success(self):
        if not self.cyresult_uptr:
            return None
        return self.cyresult_uptr.get().success
        
    @property
    def message(self):
        if not self.cyresult_uptr:
            return None
        # Cython automatically converts std::string to Python unicode string
        return self.cyresult_uptr.get().message.decode("utf-8")
    
    @property
    def integration_method(self):
        """Returns a string stating what integration method was used for this solution."""
        if not self.cyresult_uptr:
            return None
        return CyrkODEMethods.at(self.cyresult_uptr.get().integrator_method).decode("utf-8")
    
    @property
    def method(self):
        """Alias for `self.integration_method`."""
        return self.integration_method

    @property
    def t(self):
        return np.copy(np.asarray(self.time_view, dtype=np.float64, order='C'))
    
    @property
    def y(self):
        return np.copy(np.asarray(self.y_view, dtype=np.float64, order='C')).reshape((self.size, self.num_dy)).T
    
    @property
    def size(self):
        if not self.cyresult_uptr:
            return 0
        return self.cyresult_uptr.get().size
    
    @property
    def steps_taken(self):
        if not self.cyresult_uptr:
            return 0
        return self.cyresult_uptr.get().steps_taken
    
    @property
    def num_y(self):
        if not self.cyresult_uptr:
            return None
        return self.cyresult_uptr.get().num_y
    
    @property
    def num_dy(self):
        if not self.cyresult_uptr:
            return None
        return self.cyresult_uptr.get().num_dy
    
    @property
    def num_events(self):
        if not self.cyresult_uptr:
            return None
        return self.cyresult_uptr.get().num_events
    
    @property
    def t_events(self):
        if not self.cyresult_uptr:
            return None
        if self.num_events == 0:
            return None
        return self.list_t_events
    
    @property
    def y_events(self):
        if not self.cyresult_uptr:
            return None
        if self.num_events == 0:
            return None
        return self.list_y_events

    @property
    def event_terminated(self):
        if not self.cyresult_uptr:
            return None
        return self.cyresult_uptr.get().event_terminated
    
    @property
    def event_terminate_index(self):
        if not self.cyresult_uptr:
            return None
        return self.cyresult_uptr.get().event_terminate_index
    
    @property
    def status(self):
        if not self.cyresult_uptr:
            return None
        return self.cyresult_uptr.get().status
    
    @property
    def error_code(self):
        return self.status
    
    @property
    def status_message(self):
        return CyrkErrorMessages.at(self.cyresult_uptr.get().status).decode("utf-8")
    
    def __call__(self, t):

        if not self.cyresult_uptr:
            return None

        if type(t) == cnp.ndarray:
            return self.call_vectorize(t)
        else:
            return self.call(t)


# =====================================================================================================================
# Create Wrapped cysolve_ivp (has various defaults)
# =====================================================================================================================
cdef void cysolve_ivp_noreturn(
            CySolverResult* solution_ptr,
            DiffeqFuncType diffeq_ptr,
            const double t_start,
            const double t_end,
            vector[double] y0_vec,
            double rtol = 1.0e-3,
            double atol = 1.0e-6,
            vector[char] args_vec = vector[char](),
            size_t num_extra = 0,
            size_t max_num_steps = 0,
            size_t max_ram_MB = 2000,
            cpp_bool dense_output = False,
            vector[double] t_eval_vec = vector[double](),
            PreEvalFunc pre_eval_func = NULL,
            vector[Event] events_vec = vector[Event](),
            vector[double] rtols_vec = vector[double](),
            vector[double] atols_vec = vector[double](),
            double max_step = MAX_STEP,
            double first_step = 0.0,
            size_t expected_size = 0
            ) noexcept nogil:
    
    if rtols_vec.size() == 0:
        rtols_vec.resize(1)
        rtols_vec[0] = rtol
    if atols_vec.size() == 0:
        atols_vec.resize(1)
        atols_vec[0] = atol
        
    baseline_cysolve_ivp_noreturn(
        solution_ptr,
        diffeq_ptr,
        t_start,
        t_end,
        y0_vec,
        expected_size,
        num_extra,
        args_vec,
        max_num_steps,
        max_ram_MB,
        dense_output,
        t_eval_vec,
        pre_eval_func,
        events_vec,
        rtols_vec,
        atols_vec,
        max_step,
        first_step
        )

cdef CySolveOutput cysolve_ivp(
            DiffeqFuncType diffeq_ptr,
            const double t_start,
            const double t_end,
            vector[double] y0_vec,
            ODEMethod method = ODEMethod.RK45,
            double rtol = 1.0e-3,
            double atol = 1.0e-6,
            vector[char] args_vec = vector[char](),
            size_t num_extra = 0,
            size_t max_num_steps = 0,
            size_t max_ram_MB = 2000,
            cpp_bool dense_output = False,
            vector[double] t_eval_vec = vector[double](),
            PreEvalFunc pre_eval_func = NULL,
            vector[Event] events_vec = vector[Event](),
            vector[double] rtols_vec = vector[double](),
            vector[double] atols_vec = vector[double](),
            double max_step = MAX_STEP,
            double first_step = 0.0,
            size_t expected_size = 0
            ) noexcept nogil:

    if rtols_vec.size() == 0:
        rtols_vec.resize(1)
        rtols_vec[0] = rtol
    if atols_vec.size() == 0:
        atols_vec.resize(1)
        atols_vec[0] = atol
    
    cdef CySolveOutput result = baseline_cysolve_ivp(
        diffeq_ptr,
        t_start,
        t_end,
        y0_vec,
        method,
        expected_size,
        num_extra,
        args_vec,
        max_num_steps,
        max_ram_MB,
        dense_output,
        t_eval_vec,
        pre_eval_func,
        events_vec,
        rtols_vec,
        atols_vec,
        max_step,
        first_step
        )

    return move(result)

cdef CySolveOutput cysolve_ivp_gil(
            DiffeqFuncType diffeq_ptr,
            const double t_start,
            const double t_end,
            vector[double] y0_vec,
            ODEMethod method = ODEMethod.RK45,
            double rtol = 1.0e-3,
            double atol = 1.0e-6,
            vector[char] args_vec = vector[char](),
            size_t num_extra = 0,
            size_t max_num_steps = 0,
            size_t max_ram_MB = 2000,
            cpp_bool dense_output = False,
            vector[double] t_eval_vec = vector[double](),
            PreEvalFunc pre_eval_func = NULL,
            vector[Event] events_vec = vector[Event](),
            vector[double] rtols_vec = vector[double](),
            vector[double] atols_vec = vector[double](),
            double max_step = MAX_STEP,
            double first_step = 0.0,
            size_t expected_size = 0
            ) noexcept:
    
    if rtols_vec.size() == 0:
        rtols_vec.resize(1)
        rtols_vec[0] = rtol
    if atols_vec.size() == 0:
        atols_vec.resize(1)
        atols_vec[0] = atol

    cdef CySolveOutput result = baseline_cysolve_ivp(
        diffeq_ptr,
        t_start,
        t_end,
        y0_vec,
        method,
        expected_size,
        num_extra,
        args_vec,
        max_num_steps,
        max_ram_MB,
        dense_output,
        t_eval_vec,
        pre_eval_func,
        events_vec,
        rtols_vec,
        atols_vec,
        max_step,
        first_step
        )

    return move(result)



# distutils: language = c++
# cython: boundscheck=False, wraparound=False, nonecheck=False, cdivision=True, initializedcheck=False

# The "public api" prefix tells Cython to produce header files "pysolver_api.h" which can be included in C++ files.

cdef public api void call_diffeq_from_cython(object py_instance, DiffeqMethod diffeq):
    """Callback function used by the C++ model to call user-provided python diffeq functions.
    """

    # Call the python diffeq.
    diffeq(py_instance)

cdef public api double call_pyevent_from_cython(object py_instance, PyEventMethod pyevent_method, size_t event_index, double t, double* y_ptr):
    """Callback function used by the C++ model to call user-provided python event functions.
    """

    # Call the python pyevent function.
    return pyevent_method(py_instance, event_index, t, y_ptr)



from libcpp cimport bool as cpp_bool
from libcpp.vector cimport vector
from libcpp.memory cimport unique_ptr, make_unique
from libcpp.string cimport string as cpp_string
from libcpp.map cimport map as cpp_map

cimport cpython.ref as cpy_ref
from CyRK.cy.common cimport CyrkErrorCodes, CyrkErrorMessages, INF, EPS_100, BUFFER_SIZE, MAX_STEP, PreEvalFunc, DiffeqFuncType, round_to_2, find_expected_size

from CyRK.cy.pysolver_cyhook cimport DiffeqMethod
from CyRK.cy.events cimport Event

cimport numpy as cnp
cnp.import_array()


cdef extern from "c_brentq.cpp" nogil:
    pass

cdef extern from "c_events.cpp" nogil:
    pass

cdef extern from "cy_array.cpp" nogil:
    size_t binary_search_with_guess(double key, const double* array, size_t length, size_t guess)


cdef extern from "dense.cpp" nogil:
    cdef cppclass CySolverDense:
        CySolverDense()
        CySolverDense(
            CySolverBase* solver_ptr,
            cpp_bool set_state)

        void set_state()
        void call(double t_interp, double* y_interped)


# =====================================================================================================================
# Import CySolverResult (container for integration results)
# =====================================================================================================================
cdef extern from "cysolution.cpp" nogil:
    cdef cppclass CySolverResult:
            CySolverResult()
            CySolverResult(ODEMethod integration_method_)
            
            cpp_string message
            unique_ptr[ProblemConfig] config_uptr
            ODEMethod integrator_method
            CyrkErrorCodes status
            size_t size
            size_t num_interpolates
            size_t steps_taken
            cpp_bool setup_called
            cpp_bool success
            cpp_bool retain_solver
            cpp_bool event_terminated
            cpp_bool capture_dense_output
            cpp_bool capture_extra
            cpp_bool t_eval_provided
            cpp_bool direction_flag
            size_t num_events
            size_t event_terminate_index
            size_t num_y
            size_t num_dy
            vector[vector[double]] event_times
            vector[vector[double]] event_states
            vector[double] time_domain_vec
            vector[double] solution
            vector[double] time_domain_vec_sorted
            vector[double]* time_domain_vec_sorted_ptr
            vector[CySolverDense] dense_vec
            unique_ptr[CySolverBase] solver_uptr
            vector[double] interp_time_vec

            void update_status(CyrkErrorCodes status_code)
            CyrkErrorCodes setup()
            CyrkErrorCodes setup(ProblemConfig* config_ptr)
            void save_data(double new_t, double* new_solution_y, double* new_solution_dy) noexcept
            void save_event_data(size_t event_index, double event_t, double* event_y_ptr, double* event_dy_ptr) noexcept
            void record_event_data() noexcept
            void build_dense(cpp_bool save) noexcept
            CyrkErrorCodes solve()
            CyrkErrorCodes call(const double t, double* y_interp)
            CyrkErrorCodes call_vectorize(const double* t_array_ptr, size_t len_t, double* y_interp)

ctypedef unique_ptr[CySolverResult] CySolveOutput

cdef class WrapCySolverResult:
    """ Wrapper for the C++ class `CySolverResult` defined in "cysolution.cpp" """

    cdef CySolveOutput cyresult_uptr
    cdef double* time_ptr
    cdef double* y_ptr
    cdef double[::1] time_view
    cdef double[::1] y_view

    cdef list list_t_events
    cdef list list_y_events

    cdef void build_cyresult(self, ODEMethod integrator_method)
    cdef void set_cyresult_pointer(self, CySolveOutput cyresult_uptr_)
    cdef set_problem_config(self, ProblemConfig* new_problem_config_ptr)
    cpdef solve(self)
    cpdef finalize(self)

# =====================================================================================================================
# Import CySolver Integrator Base Class
# =====================================================================================================================
cdef extern from "cysolver.cpp" nogil:
    cpdef enum class ODEMethod(int):
        NO_METHOD_SET,
        BASE_METHOD,
        RK_BASE_METHOD,
        RK23,
        RK45,
        DOP853
    const cpp_map[ODEMethod, cpp_string] CyrkODEMethods

    cdef cppclass ProblemConfig:
        ProblemConfig()
        ProblemConfig(
            DiffeqFuncType diffeq_ptr_,
            double t_start_,
            double t_end_,
            vector[double]& y0_vec_)
        ProblemConfig(
            DiffeqFuncType diffeq_ptr_,
            double t_start_,
            double t_end_,
            vector[double]& y0_vec_,
            vector[char]& args_vec_,
            vector[double]& t_eval_vec_,
            size_t num_extra_,
            size_t expected_size_,
            size_t max_num_steps_,
            size_t max_ram_MB_,
            PreEvalFunc pre_eval_func_,
            cpp_bool capture_dense_output_,
            cpp_bool force_retain_solver_,
            vector[Event]& events_vec_)
        
        DiffeqFuncType diffeq_ptr
        double t_start
        double t_end
        vector[double] y0_vec
        vector[char] args_vec
        vector[double] t_eval_vec
        size_t num_extra
        size_t expected_size
        size_t max_num_steps
        size_t max_ram_MB
        PreEvalFunc pre_eval_func
        cpp_bool capture_dense_output
        cpp_bool force_retain_solver
        cpp_bool capture_extra
        cpp_bool t_eval_provided
        size_t num_y
        double num_y_dbl
        double num_y_sqrt
        size_t num_dy
        double num_dy_dbl
        cpy_ref.PyObject* cython_extension_class_instance
        DiffeqMethod py_diffeq_method
        cpp_bool initialized
        cpp_bool check_events
        vector[Event] events_vec

        void update_properties(
            DiffeqFuncType diffeq_ptr_,
            double t_start_,
            double t_end_,
            vector[double]& y0_vec_
        )
        void update_properties(
            DiffeqFuncType diffeq_ptr_,
            double t_start_,
            double t_end_,
            vector[double]& y0_vec_,
            vector[char]& args_vec_,
            vector[double]& t_eval_vec_,
            size_t num_extra_,
            size_t expected_size_,
            size_t max_num_steps_,
            size_t max_ram_MB_,
            PreEvalFunc pre_eval_func_,
            cpp_bool capture_dense_output_,
            cpp_bool force_retain_solver_,
            vector[Event]& events_vec_
        )
        void initialize()
        void update_properties_from_config(ProblemConfig* new_config_ptr)

    struct NowStatePointers:
        double* t_now_ptr
        double* y_now_ptr
        double* dy_now_ptr

    cdef cppclass CySolverBase:
        CySolverBase()
        CySolverBase(CySolverResult* storage_ptr_)

        cpp_bool use_dense_output
        cpp_bool user_provided_max_num_steps
        cpp_bool use_pysolver
        ODEMethod integration_method
        size_t num_y
        size_t num_extra
        size_t num_dy
        size_t num_events
        double t_old
        double t_now
        double* y_old_ptr
        double* y_now_ptr
        double* dy_now_ptr
        vector[size_t] active_event_indices_vec

        void set_Q_order(size_t* Q_order_ptr)
        void set_Q_array(double* Q_ptr)
        void clear_python_refs()
        void offload_to_temp() noexcept
        void load_back_from_temp() noexcept
        CyrkErrorCodes resize_num_y(size_t num_y_, size_t num_dy_)
        CyrkErrorCodes setup()
        cpp_bool check_status()
        void take_step()
        void solve()
        NowStatePointers get_now_state()
        void diffeq()
        CyrkErrorCodes set_cython_extension_instance(
            cpy_ref.PyObject* cython_extension_class_instance,
            DiffeqMethod py_diffeq_method
            )
        void py_diffeq()


# =====================================================================================================================
# Import CySolver Runge-Kutta Integrators
# =====================================================================================================================
cdef extern from "rk.cpp" nogil:

    cdef cppclass RKConfig(ProblemConfig):
        RKConfig(
            DiffeqFuncType diffeq_ptr_,
            double t_start_,
            double t_end_,
            vector[double]& y0_vec_,
            vector[char]& args_vec_,
            vector[double]& t_eval_vec_,
            size_t num_extra_,
            size_t expected_size_,
            size_t max_num_steps_,
            size_t max_ram_MB_,
            PreEvalFunc pre_eval_func_,
            cpp_bool capture_dense_output_,
            cpp_bool force_retain_solver_,
            vector[Event]& events_vec_,
            vector[double]& rtols_,
            vector[double]& atols_,
            double max_step_size_,
            double first_step_size_)
        
        vector[double] rtols
        vector[double] atols
        double max_step_size
        double first_step_size

        void update_properties(
            DiffeqFuncType diffeq_ptr_,
            double t_start_,
            double t_end_,
            vector[double]& y0_vec_,
            vector[char]& args_vec_,
            vector[double]& t_eval_vec_,
            size_t num_extra_,
            size_t expected_size_,
            size_t max_num_steps_,
            size_t max_ram_MB_,
            PreEvalFunc pre_eval_func_,
            cpp_bool capture_dense_output_,
            cpp_bool force_retain_solver_,
            vector[Event]& events_vec_,
            vector[double]& rtols_,
            vector[double]& atols_,
            double max_step_size_,
            double first_step_size_)
        void initialize()
        void update_properties_from_config(RKConfig* new_config_ptr)


    cdef cppclass RKSolver(CySolverBase):
        RKSolver()
        RKSolver(CySolverResult* storage_ptr_)
        void set_Q_order(size_t* Q_order_ptr)
        void set_Q_array(double* Q_ptr)
        CyrkErrorCodes setup()

    cdef cppclass RK23(RKSolver):
        RK23()
        RK23(CySolverResult* storage_ptr_)
    
    cdef cppclass RK45(RKSolver):
        RK45()
        RK45(CySolverResult* storage_ptr_)

    cdef cppclass DOP853(RKSolver):
        DOP853()
        DOP853(CySolverResult* storage_ptr_)

# =====================================================================================================================
# Import the C++ cysolve_ivp helper function
# =====================================================================================================================
cdef extern from "cysolve.cpp" nogil:
    # Pure C++ and Cython implementation

    cdef void baseline_cysolve_ivp_noreturn(
        CySolverResult* solution_ptr,
        DiffeqFuncType diffeq_ptr,
        double t_start,
        double t_end,
        vector[double]& y0_vec,
        size_t expected_size,
        size_t num_extra,
        vector[char]& args_vec,
        size_t max_num_steps,
        size_t max_ram_MB,
        cpp_bool capture_dense_output,
        vector[double]& t_eval_vec,
        PreEvalFunc pre_eval_func,
        vector[Event]& events_vec,
        vector[double]& rtols,
        vector[double]& atols,
        double max_step_size,
        double first_step_size
        )

    cdef unique_ptr[CySolverResult] baseline_cysolve_ivp(
        DiffeqFuncType diffeq_ptr,
        double t_start,
        double t_end,
        vector[double]& y0_vec,
        ODEMethod integration_method,
        size_t expected_size,
        size_t num_extra,
        vector[char]& args_vec,
        size_t max_num_steps,
        size_t max_ram_MB,
        cpp_bool capture_dense_output,
        vector[double]& t_eval_vec,
        PreEvalFunc pre_eval_func,
        vector[Event]& events_vec,
        vector[double]& rtols,
        vector[double]& atols,
        double max_step_size,
        double first_step_size
        )


# =====================================================================================================================
# Cython-based wrapper for baseline_cysolve_ivp that carries default values.
# =====================================================================================================================
cdef void cysolve_ivp_noreturn(
    CySolverResult* solution_ptr,
    DiffeqFuncType diffeq_ptr,
    const double t_start,
    const double t_end,
    vector[double] y0_vec,
    double rtol = *,
    double atol = *,
    vector[char] args_vec = *,
    size_t num_extra = *,
    size_t max_num_steps = *,
    size_t max_ram_MB = *,
    bint dense_output = *,
    vector[double] t_eval_vec = *,
    PreEvalFunc pre_eval_func = *,
    vector[Event] events_vec = *,
    vector[double] rtols_vec = *,
    vector[double] atols_vec = *,
    double max_step = *,
    double first_step = *,
    size_t expected_size = *
    ) noexcept nogil

cdef CySolveOutput cysolve_ivp(
    DiffeqFuncType diffeq_ptr,
    const double t_start,
    const double t_end,
    vector[double] y0_vec,
    ODEMethod method = *,
    double rtol = *,
    double atol = *,
    vector[char] args_vec = *,
    size_t num_extra = *,
    size_t max_num_steps = *,
    size_t max_ram_MB = *,
    bint dense_output = *,
    vector[double] t_eval_vec = *,
    PreEvalFunc pre_eval_func = *,
    vector[Event] events_vec = *,
    vector[double] rtols_vec = *,
    vector[double] atols_vec = *,
    double max_step = *,
    double first_step = *,
    size_t expected_size = *
    ) noexcept nogil

cdef CySolveOutput cysolve_ivp_gil(
    DiffeqFuncType diffeq_ptr,
    const double t_start,
    const double t_end,
    vector[double] y0_vec,
    ODEMethod method = *,
    double rtol = *,
    double atol = *,
    vector[char] args_vec = *,
    size_t num_extra = *,
    size_t max_num_steps = *,
    size_t max_ram_MB = *,
    bint dense_output = *,
    vector[double] t_eval_vec = *,
    PreEvalFunc pre_eval_func = *,
    vector[Event] events_vec = *,
    vector[double] rtols_vec = *,
    vector[double] atols_vec = *,
    double max_step = *,
    double first_step = *,
    size_t expected_size = *
    ) noexcept



import numpy as np
import pytest


def test_build_events():
    """Basic tests if the C++ event wrapper class, further wrapped by cython, can be built and initialized correctly."""
    from CyRK.cy.events_test import build_event_wrapper_test

    assert build_event_wrapper_test()

@pytest.mark.parametrize('use_dense', (True, False))
@pytest.mark.parametrize('use_t_eval', (True, False))
@pytest.mark.parametrize('use_termination', (True, False))
@pytest.mark.parametrize('use_capture_extra', (True, False))
def test_run_cysolver_with_events(use_dense, use_t_eval, use_termination, use_capture_extra):
    """Test if the solver runs correctly with events."""
    from CyRK.cy.events_test import run_cysolver_with_events

    if use_t_eval:
        t_eval = np.linspace(0, 10, 11, dtype=np.float64)
    else:
        t_eval = np.array([], dtype=np.float64)

    # Run cysolver with events
    solution, cython_tests_passed = run_cysolver_with_events(use_dense, t_eval, use_termination, use_capture_extra)
    assert cython_tests_passed

    assert type(solution.t_events) == list
    assert len(solution.t_events) == 3
    assert type(solution.y_events) == list
    assert len(solution.y_events) == 3
    assert solution.event_terminated == use_termination
    if use_termination:
        assert solution.event_terminate_index == 0
    
    # Loop through 3 events
    for i in range(3):
        if i == 0 and use_termination:
            assert solution.t_events[i].size == 1

        if use_capture_extra:
            assert solution.y_events[i].shape[0] == 6  # 3 dependent y values + 3 extras
        else:
            assert solution.y_events[i].shape[0] == 3  # 3 dependent y values
        assert solution.t_events[i].size == solution.y_events[i][0].size

        if use_capture_extra:
            # The first extra output (4th output including dependent ys) is always a constant in this example
            assert np.all(solution.y_events[i][3] == 10.0)
    
    if use_termination:
        assert solution.t[-1] < 5.1


if __name__ == "__main__":
    for i in range(100_000):
        print(i)
        test_run_cysolver_with_events(False, False, True, True)


from CyRK.cy.common cimport PreEvalFunc

cdef void baseline_diffeq(double* dy_ptr, double t, double* y_ptr, char* args_ptr, PreEvalFunc pre_eval_func) noexcept nogil
cdef void accuracy_test_diffeq(double* dy_ptr, double t, double* y_ptr, char* args_ptr, PreEvalFunc pre_eval_func) noexcept nogil
cdef void extraoutput_test_diffeq(double* dy_ptr, double t, double* y_ptr, char* args_ptr, PreEvalFunc pre_eval_func) noexcept nogil
cdef void lorenz_diffeq(double* dy_ptr, double t, double* y_ptr, char* args_ptr, PreEvalFunc pre_eval_func) noexcept nogil
cdef void lorenz_extraoutput_diffeq(double* dy_ptr, double t, double* y_ptr, char* args_ptr, PreEvalFunc pre_eval_func) noexcept nogil
cdef void lotkavolterra_diffeq(double* dy_ptr, double t, double* y_ptr, char* args_ptr, PreEvalFunc pre_eval_func) noexcept nogil
cdef void pendulum_diffeq(double* dy_ptr, double t, double* y_ptr, char* args_ptr, PreEvalFunc pre_eval_func) noexcept nogil



from libcpp cimport bool as cpp_bool
from libcpp.string cimport string as cpp_string
from libcpp.map cimport map as cpp_map
from libcpp.vector cimport vector

# =====================================================================================================================
# Import common functions and constants
# =====================================================================================================================
cdef extern from "c_common.cpp" nogil:
    cpdef enum class CyrkErrorCodes(int):
        CONVERGED,
        INITIALIZING,
        EVENT_TERMINATED,
        SUCCESSFUL_INTEGRATION,
        NO_ERROR,
        GENERAL_ERROR,
        PROPERTY_NOT_SET,
        UNSUPPORTED_UNKNOWN_MODEL,
        UNINITIALIZED_CLASS,
        CYSOLVER_INITIALIZATION_ERROR,
        INCOMPATIBLE_INPUT,
        ATTRIBUTE_ERROR,
        BOUNDS_ERROR,
        ARGUMENT_NOT_SET,
        ARGUMENT_ERROR,
        SETUP_NOT_CALLED,
        DENSE_OUTPUT_NOT_SAVED,
        BAD_CONFIG_DATA,
        OPTIMIZE_SIGN_ERROR,
        OPTIMIZE_CONVERGENCE_ERROR,
        MEMORY_ALLOCATION_ERROR,
        VECTOR_SIZE_EXCEEDS_LIMITS,
        NUMBER_OF_EQUATIONS_IS_ZERO,
        MAX_ITERATIONS_HIT,
        MAX_STEPS_USER_EXCEEDED,
        MAX_STEPS_SYSARCH_EXCEEDED,
        STEP_SIZE_ERROR_SPACING,
        STEP_SIZE_ERROR_ACCEPTANCE,
        DENSE_BUILD_FAILED,
        INTEGRATION_NOT_SUCCESSFUL,
        EVENT_SETUP_FAILED,
        ERROR_IMPORTING_PYTHON_MODULE,
        BAD_INITIAL_STEP_SIZE,
        OTHER_ERROR,
        UNSET_ERROR_CODE
    
    const cpp_map[CyrkErrorCodes, cpp_string] CyrkErrorMessages

    cdef struct OptimizeInfo:
        size_t funcalls
        size_t iterations
        CyrkErrorCodes error_num
        vector[double] y_vec
        double* y_at_root_ptr
    
    const double INF
    const double EPS_100
    const size_t BUFFER_SIZE
    const double MAX_STEP
    const double dbl_NAN
    const size_t MAX_SIZET_SIZE

    ctypedef void (*PreEvalFunc)(char*, double, double*, char*)
    ctypedef void (*DiffeqFuncType)(double*, double, double*, char*, PreEvalFunc)

    cdef void round_to_2(size_t& initial_value) noexcept

    cdef size_t find_expected_size(
        size_t num_y,
        size_t num_extra,
        double t_delta_abs,
        double rtol_min)



# distutils: language = c++
# cython: boundscheck=False, wraparound=False, nonecheck=False, cdivision=True, initializedcheck=False
from libc.math cimport sin, cos
from libc.string cimport memcpy

from libcpp cimport bool as cpp_bool
from libcpp.utility cimport move
from libcpp.limits cimport numeric_limits
from libcpp.vector cimport vector

cdef double d_NAN = numeric_limits[double].quiet_NaN()

from CyRK.cy.common cimport DiffeqFuncType, MAX_STEP
from CyRK.cy.cysolver_api cimport cysolve_ivp_noreturn, cysolve_ivp, WrapCySolverResult, CySolveOutput, ODEMethod, CySolverResult
from CyRK.cy.events cimport Event

import numpy as np
cimport numpy as np
np.import_array()


cdef void baseline_diffeq(double* dy_ptr, double t, double* y_ptr, char* args_ptr, PreEvalFunc pre_eval_func) noexcept nogil:
    # Unpack y
    cdef double y0, y1
    y0 = y_ptr[0]
    y1 = y_ptr[1]

    dy_ptr[0] = (1. - 0.01 * y1) * y0
    dy_ptr[1] = (0.02 * y0 - 1.) * y1


cdef void accuracy_test_diffeq(double* dy_ptr, double t, double* y_ptr, char* args_ptr, PreEvalFunc pre_eval_func) noexcept nogil:
    # Unpack y
    cdef double y0, y1
    y0 = y_ptr[0]
    y1 = y_ptr[1]

    dy_ptr[0] = sin(t) - y1
    dy_ptr[1] = cos(t) + y0


cdef void extraoutput_test_diffeq(double* dy_ptr, double t, double* y_ptr, char* args_ptr, PreEvalFunc pre_eval_func) noexcept nogil:
    # Unpack y
    cdef double y0, y1, extra_0, extra_1
    y0 = y_ptr[0]
    y1 = y_ptr[1]

    extra_0 = (1. - 0.01 * y1)
    extra_1 = (0.02 * y0 - 1.)

    # Store dy/dt
    dy_ptr[0] = extra_0 * y0
    dy_ptr[1] = extra_1 * y1

    # Store extra output
    dy_ptr[2] = extra_0
    dy_ptr[3] = extra_1


cdef void lorenz_diffeq(double* dy_ptr, double t, double* y_ptr, char* args_ptr, PreEvalFunc pre_eval_func) noexcept nogil:
    # Unpack args
    cdef double* args_dbl_ptr = <double*>args_ptr
    cdef double a = args_dbl_ptr[0]
    cdef double b = args_dbl_ptr[1]
    cdef double c = args_dbl_ptr[2]

    # Unpack y
    cdef double y0, y1, y2
    y0 = y_ptr[0]
    y1 = y_ptr[1]
    y2 = y_ptr[2]

    dy_ptr[0] = a * (y1 - y0)
    dy_ptr[1] = y0 * (b - y2) - y1
    dy_ptr[2] = y0 * y1 - c * y2


cdef void lorenz_extraoutput_diffeq(double* dy_ptr, double t, double* y_ptr, char* args_ptr, PreEvalFunc pre_eval_func) noexcept nogil:
    # Unpack args
    cdef double* args_dbl_ptr = <double*>args_ptr
    cdef double a = args_dbl_ptr[0]
    cdef double b = args_dbl_ptr[1]
    cdef double c = args_dbl_ptr[2]

    # Unpack y
    cdef double y0, y1, y2
    y0 = y_ptr[0]
    y1 = y_ptr[1]
    y2 = y_ptr[2]

    cdef double e_1 = a
    cdef double e_2 = (b - y2)
    cdef double e_3 = c * y2

    dy_ptr[0] = e_1 * (y1 - y0)
    dy_ptr[1] = y0 * e_2 - y1
    dy_ptr[2] = y0 * y1 - e_3

    dy_ptr[3] = e_1
    dy_ptr[4] = e_2
    dy_ptr[5] = e_3


cdef void lotkavolterra_diffeq(double* dy_ptr, double t, double* y_ptr, char* args_ptr, PreEvalFunc pre_eval_func) noexcept nogil:
    # Unpack args
    cdef double* args_dbl_ptr = <double*>args_ptr
    cdef double a = args_dbl_ptr[0]
    cdef double b = args_dbl_ptr[1]
    cdef double c = args_dbl_ptr[2]
    cdef double d = args_dbl_ptr[3]

    # Unpack y
    cdef double y0, y1
    y0 = y_ptr[0]
    y1 = y_ptr[1]

    dy_ptr[0] = a * y0 - b * y0 * y1
    dy_ptr[1] = -c * y1 + d * y0 * y1



cdef void pendulum_diffeq(double* dy_ptr, double t, double* y_ptr, char* args_ptr, PreEvalFunc pre_eval_func) noexcept nogil:
    # Unpack args
    cdef double* args_dbl_ptr = <double*>args_ptr
    cdef double l = args_dbl_ptr[0]
    cdef double m = args_dbl_ptr[1]
    cdef double g = args_dbl_ptr[2]

    cdef double coeff_1 = (-3. * g / (2. * l))
    cdef double coeff_2 = (3. / (m * l**2))

    # Unpack y
    cdef double y0, y1, torque
    y0 = y_ptr[0]
    y1 = y_ptr[1]

    # External torque
    torque = 0.1 * sin(t)

    dy_ptr[0] = y1
    dy_ptr[1] = coeff_1 * sin(y0) + coeff_2 * torque


cdef struct ArbitraryArgStruct:
    double l
    # Let's make sure it has something in the middle that is not a double for size checks.
    cpp_bool cause_fail
    cpp_bool checker
    double m
    double g

cdef void arbitrary_arg_test(double* dy_ptr, double t, double* y_ptr, char* args_ptr, PreEvalFunc pre_eval_func) noexcept nogil:
    # Unpack args
    cdef ArbitraryArgStruct* arb_args_ptr = <ArbitraryArgStruct*>args_ptr
    cdef double l = arb_args_ptr.l
    cdef double m = arb_args_ptr.m
    cdef double g = arb_args_ptr.g
    cdef cpp_bool cause_fail = arb_args_ptr.cause_fail
    cdef cpp_bool checker    = arb_args_ptr.checker

    cdef double coeff_1 = (-3. * g / (2. * l))
    cdef double coeff_2 = (3. / (m * l**2))

    # Unpack y
    cdef double y0, y1, torque
    y0 = y_ptr[0]
    y1 = y_ptr[1]

    # External torque
    torque = 0.1 * sin(t)

    if cause_fail:
        # Put something crazy that will cause problems
        dy_ptr[0] = (100 * t)**(1000)
        dy_ptr[1] = (-100 * t)**(1000)
    else:
        dy_ptr[0] = y1
        dy_ptr[1] = coeff_1 * sin(y0) + coeff_2 * torque



cdef void pendulum_preeval_func(char* output_ptr, double time, double* y_ptr, char* args_ptr) noexcept nogil:

    # Unpack args
    cdef double* args_dbl_ptr = <double*>args_ptr
    cdef double l = args_dbl_ptr[0]
    cdef double m = args_dbl_ptr[1]
    cdef double g = args_dbl_ptr[2]

    cdef double coeff_1 = (-3. * g / (2. * l))
    cdef double coeff_2 = (3. / (m * l**2))

    # Unpack y
    cdef double torque

    # External torque
    torque = 0.1 * sin(time)

    # Convert output pointer to double pointer so we can store data
    cdef double* output_dbl_ptr = <double*>output_ptr
    output_dbl_ptr[0] = torque
    output_dbl_ptr[1] = coeff_1
    output_dbl_ptr[2] = coeff_2


cdef void pendulum_preeval_diffeq(double* dy_ptr, double t, double* y_ptr, char* args_ptr, PreEvalFunc pre_eval_func) noexcept nogil:
    
    # Make stack allocated storage for pre eval output
    cdef double[3] pre_eval_storage
    cdef double* pre_eval_storage_ptr = &pre_eval_storage[0]

    # Cast storage to void so we can call function
    cdef char* pre_eval_storage_char_ptr = <char*>pre_eval_storage_ptr

    # Call Pre-Eval Function
    pre_eval_func(pre_eval_storage_char_ptr, t, y_ptr, args_ptr)

    cdef double y0 = y_ptr[0]
    cdef double y1 = y_ptr[1]

    dy_ptr[0] = y1
    dy_ptr[1] = pre_eval_storage_ptr[1] * sin(y0) + pre_eval_storage_ptr[2] * pre_eval_storage_ptr[0]

def cy_extra_output_tester():

    cdef double t_start = 0.0
    cdef double t_end = 10.0

    cdef vector[double] y0_vec = vector[double](3)
    y0_vec[0] = 1.0
    y0_vec[1] = 0.0
    y0_vec[2] = 0.0
    
    cdef size_t num_y     = 3
    cdef size_t num_extra = 3
    cdef ODEMethod int_method = ODEMethod.RK45

    cdef size_t arg_size = sizeof(double) * 3
    cdef vector[char] args_vec = vector[char](arg_size)
    cdef double* args_ptr = <double*>args_vec.data()
    args_ptr[0] = 10.0
    args_ptr[1] = 28.0
    args_ptr[2] = 8.0 / 3.0

    cdef CySolveOutput result = cysolve_ivp(
        lorenz_extraoutput_diffeq,
        t_start,
        t_end,
        y0_vec,
        method=int_method,
        rtol=1.0e-4,
        atol=1.0e-5,
        args_vec=args_vec,
        num_extra=num_extra,
        max_num_steps=0,
        max_ram_MB=2000,
        dense_output=True
        )

    cdef double check_t = 4.335
    assert result.get().success

    # Call the result to get the baseline values for extra output
    cdef double[6] y_interp
    cdef double* y_interp_ptr = &y_interp[0]
    result.get().call(check_t, y_interp_ptr)
    cdef double dy1 = y_interp_ptr[0]
    cdef double dy2 = y_interp_ptr[1]
    cdef double dy3 = y_interp_ptr[2]
    cdef double e1  = y_interp_ptr[3]
    cdef double e2  = y_interp_ptr[4]
    cdef double e3  = y_interp_ptr[5]

    # Recall the solution to see if it still produces the expected values
    result.get().call(check_t, y_interp_ptr)

    np.testing.assert_allclose(dy1, y_interp_ptr[0])
    np.testing.assert_allclose(dy2, y_interp_ptr[1])
    np.testing.assert_allclose(dy3, y_interp_ptr[2])
    np.testing.assert_allclose(e1, y_interp_ptr[3])
    np.testing.assert_allclose(e2, y_interp_ptr[4])
    np.testing.assert_allclose(e3, y_interp_ptr[5])

    return True

# Event fuinctions for baseline diffeq
cdef double baseline_diffeq_event1_check(double t, double* y, char* args) noexcept nogil:
    if y[0] > 50.0:
        return 0.0
    return 1.0

cdef double baseline_diffeq_event2_check(double t, double* y, char* args) noexcept nogil:
    if y[1] < 50.0:
        return 0.0
    return 1.0

cdef double baseline_diffeq_event3_check(double t, double* y, char* args) noexcept nogil:
    if t > 5.0:
        return 0.0
    else:
        return 1.0

cdef vector[Event] baseline_diffeq_events = vector[Event]()
# Default constructor will handle direction and max num allowed (no termination in this case)
baseline_diffeq_events.emplace_back(baseline_diffeq_event1_check)
baseline_diffeq_events.emplace_back(baseline_diffeq_event2_check)
baseline_diffeq_events.emplace_back(baseline_diffeq_event3_check)


# Event functions for pendulum diffeq
cdef double pendulum_diffeq_event1_check(double t, double* y, char* args) noexcept nogil:
    if y[0] > 50.0:
        return 0.0
    return 1.0

cdef double pendulum_diffeq_event2_check(double t, double* y, char* args) noexcept nogil:
    if y[1] < 50.0:
        return 0.0
    return 1.0

cdef double pendulum_diffeq_event3_check(double t, double* y, char* args) noexcept nogil:
    if t > 5.0:
        return 0.0
    else:
        return 1.0

cdef vector[Event] pendulum_diffeq_events = vector[Event]()
# Default constructor will handle direction and max num allowed (no termination in this case)
pendulum_diffeq_events.emplace_back(pendulum_diffeq_event1_check)
pendulum_diffeq_events.emplace_back(pendulum_diffeq_event2_check)
pendulum_diffeq_events.emplace_back(pendulum_diffeq_event3_check)


def cytester(
        int diffeq_number,
        tuple t_span = None,
        const double[::1] y0 = None,
        double[::1] args = None,
        str method = 'rk45',
        size_t expected_size = 0,
        size_t max_num_steps = 0,
        size_t max_ram_MB = 2000,
        bint dense_output = False,
        bint use_events = False,
        double[::1] t_eval = None,
        double rtol = 1.0e-3, 
        double atol = 1.0e-6,
        double[::1] rtol_array = None,
        double[::1] atol_array = None,
        double max_step = MAX_STEP,
        double first_step = 0.0,
        WrapCySolverResult solution_reuse = None
        ):
    cdef size_t i
    cdef vector[double] t_eval_vec = vector[double]()
    if t_eval is not None:
        for i in range(t_eval.size):
            t_eval_vec.push_back(t_eval[i])

    cdef size_t num_y
    cdef vector[double] y0_vec = vector[double]()

    cdef double t_start = 0.0
    cdef double t_end   = 0.0

    # Parse method
    method = method.lower()
    cdef ODEMethod integration_method = ODEMethod.RK45
    if method == "rk23":
        integration_method = ODEMethod.RK23
    elif method == "rk45":
        integration_method = ODEMethod.RK45
    elif method == 'dop853':
        integration_method = ODEMethod.DOP853
    else:
        raise NotImplementedError(
            "ERROR: `PySolver::set_problem_parameters` - "
            f"Unknown or unsupported integration method provided: {method}.\n"
            f"Supported methods are: RK23, RK45, DOP853."
            )

    cdef size_t num_extra = 0
    cdef DiffeqFuncType diffeq = NULL
    cdef PreEvalFunc pre_eval_func = NULL
    if diffeq_number == 0:
        diffeq = baseline_diffeq
    elif diffeq_number == 1:
        diffeq = accuracy_test_diffeq
    elif diffeq_number == 2:
        diffeq = extraoutput_test_diffeq
        num_extra = 2
    elif diffeq_number == 3:
        diffeq = lorenz_diffeq
    elif diffeq_number == 4:
        diffeq = lorenz_extraoutput_diffeq
        num_extra = 3
    elif diffeq_number == 5:
        diffeq = lotkavolterra_diffeq
    elif diffeq_number == 6:
        diffeq = pendulum_diffeq
    elif diffeq_number == 7:
        diffeq = arbitrary_arg_test
    elif diffeq_number == 8:
        diffeq = pendulum_preeval_diffeq
        pre_eval_func = pendulum_preeval_func
    else:
        raise NotImplementedError
    
    cdef vector[Event] events_vec = vector[Event]()
    if use_events:
        if diffeq_number == 0:
            events_vec = baseline_diffeq_events
        elif diffeq_number == 6:
            events_vec = pendulum_diffeq_events
        else:
            raise NotImplementedError

    # Set up additional argument information
    cdef vector[char] args_vec = vector[char]()
    cdef size_t size_of_args = 0
    cdef double* args_dbl_ptr = NULL
    # Abitrary arg test requires a ArbitraryArgStruct class instance to be passed in
    cdef ArbitraryArgStruct arb_arg_struct = ArbitraryArgStruct(1.0, False, True, 1.0, 9.81)
    
    # Check if generic testing was requested.
    if y0 is None:
        # Generic Testing Requested
        
        # Use a higher tolerance for testing
        rtol = 1.0e-6
        atol = 1.0e-8

        # Get inputs for requested functions
        if diffeq_number == 0:
            num_y = 2
            y0_vec.resize(num_y)
            y0_vec[0] = 20.0
            y0_vec[1] = 20.0
            t_start = 0.0
            t_end = 20.0

        elif diffeq_number == 1:
            num_y = 2
            y0_vec.resize(num_y)
            y0_vec[0] = 0.0
            y0_vec[1] = 1.0
            t_start = 0.0
            t_end = 10.0

        elif diffeq_number == 2:
            num_y = 2
            y0_vec.resize(num_y)
            y0_vec[0] = 20.0
            y0_vec[1] = 20.0
            t_start = 0.0
            t_end = 20.0

        elif diffeq_number == 3:
            num_y = 3
            y0_vec.resize(num_y)
            y0_vec[0] = 1.0
            y0_vec[1] = 0.0
            y0_vec[2] = 0.0
            t_start = 0.0
            t_end = 10.0
            args_vec.resize(3 * sizeof(double))
            args_dbl_ptr = <double*>args_vec.data()
            args_dbl_ptr[0] = 10.0
            args_dbl_ptr[1] = 28.0
            args_dbl_ptr[2] = 8.0 / 3.0
            
        elif diffeq_number == 4:
            num_y = 3
            y0_vec.resize(num_y)
            y0_vec[0] = 1.0
            y0_vec[1] = 0.0
            y0_vec[2] = 0.0
            t_start = 0.0
            t_end = 10.0
            args_vec.resize(3 * sizeof(double))
            args_dbl_ptr = <double*>args_vec.data()
            args_dbl_ptr[0] = 10.0
            args_dbl_ptr[1] = 28.0
            args_dbl_ptr[2] = 8.0 / 3.0

        elif diffeq_number == 5:
            num_y = 2
            y0_vec.resize(num_y)
            y0_vec[0] = 10.0
            y0_vec[1] = 5.0
            t_start = 0.0
            t_end = 15.0
            args_vec.resize(4 * sizeof(double))
            args_dbl_ptr = <double*>args_vec.data()
            args_dbl_ptr[0] = 1.5
            args_dbl_ptr[1] = 1.0
            args_dbl_ptr[2] = 3.0
            args_dbl_ptr[3] = 1.0

        elif diffeq_number == 6:
            num_y = 2
            y0_vec.resize(num_y)
            y0_vec[0] = 0.01
            y0_vec[1] = 0.0
            t_start = 0.0
            t_end = 10.0
            args_vec.resize(3 * sizeof(double))
            args_dbl_ptr = <double*>args_vec.data()
            args_dbl_ptr[0] = 1.0
            args_dbl_ptr[1] = 1.0
            args_dbl_ptr[2] = 9.81

        elif diffeq_number == 7:
            num_y = 2
            y0_vec.resize(num_y)
            y0_vec[0] = 0.01
            y0_vec[1] = 0.0
            t_start = 0.0
            t_end = 10.0
            # Set args pointer to our arb args struct variable's address and cast it to a void pointer
            args_vec.resize(sizeof(ArbitraryArgStruct))
            memcpy(args_vec.data(), &arb_arg_struct, sizeof(ArbitraryArgStruct))
        
        elif diffeq_number == 8:
            num_y = 2
            y0_vec.resize(num_y)
            y0_vec[0] = 0.01
            y0_vec[1] = 0.0
            t_start = 0.0
            t_end = 10.0
            args_vec.resize(3 * sizeof(double))
            args_dbl_ptr = <double*>args_vec.data()
            args_dbl_ptr[0] = 1.0
            args_dbl_ptr[1] = 1.0
            args_dbl_ptr[2] = 9.81

        else:
            raise NotImplementedError
    else:
        # Regular testing requested.
        num_y = len(y0)
        y0_vec.resize(num_y)
        for i in range(num_y):
            y0_vec[i] = y0[i]
        if t_span is not None:
            t_start = t_span[0]
            t_end   = t_span[1]
        else:
            raise AttributeError("ERROR: `cytester`: t_span not provided.")
        if args is not None:
            # This tester assumes that the args input is an array of doubles.
            args_vec.resize(args.size * sizeof(double))
            memcpy(args_vec.data(), &args[0], sizeof(double) * args.size)
        else:
            args_vec.resize(0)

    # Parse rtol
    cdef vector[double] rtols_vec = vector[double]()
    if rtol_array is not None:
        rtols_vec.resize(rtol_array.size)
        for i in range(rtol_array.size):
            rtols_vec[i] = rtol_array[i]
    
    # Parse atol
    cdef vector[double] atols_vec = vector[double]()
    if atol_array is not None:
        atols_vec.resize(atol_array.size)
        for i in range(atol_array.size):
            atols_vec[i] = atol_array[i]

    # Build python-safe solution storage
    if solution_reuse is None:
        solution_reuse = WrapCySolverResult()
        solution_reuse.build_cyresult(integration_method)
    else:
        if not solution_reuse.cyresult_uptr:
            solution_reuse.build_cyresult(integration_method)
        elif integration_method != solution_reuse.cyresult_uptr.get().integrator_method:
            raise AttributeError("Can not reuse solution that has different integration method.")

    cdef CySolverResult* solution_ptr = solution_reuse.cyresult_uptr.get()

    if not solution_ptr:
        raise AttributeError("Solution pointer not set.")

    # Solve ODE
    cysolve_ivp_noreturn(
        solution_ptr,
        diffeq,
        t_start,
        t_end,
        y0_vec,
        rtol = rtol,
        atol = atol,
        args_vec = args_vec,
        num_extra = num_extra,
        max_num_steps = max_num_steps,
        max_ram_MB = max_ram_MB,
        dense_output = dense_output,
        t_eval_vec = t_eval_vec,
        pre_eval_func = pre_eval_func,
        events_vec = events_vec,
        rtols_vec = rtols_vec,
        atols_vec = atols_vec,
        max_step = max_step,
        first_step = first_step,
        expected_size = expected_size
        )
    
    solution_reuse.finalize()

    return solution_reuse



import pytest
import numpy as np
import numba as nb
from CyRK import pysolve_ivp
import math

# define base system
y_0 = 1
t_span = (0.0, 8.0)
@nb.njit("void(f8[::1], f8, f8[::1])")
def cy_diffeq(dy, t, y):
    for i in range(y.size):
        dy[i] = -y[i]/2 + 2*math.sin(3*t)
    

@pytest.mark.parametrize('Neqs', (2**22,))
def test_pysolve_large_neqs(Neqs):
    """ Tests CyRK's ability to simultaneously solve many copies of this system. """
    y0 = np.full(Neqs, y_0, dtype=np.float64)
    # For large Neq we need to increase the max_ram_MB limit.
    result = pysolve_ivp(cy_diffeq, t_span, y0, method="RK45", rtol=1e-7, atol=1e-7, pass_dy_as_arg=True, max_ram_MB=8_000)
    assert result.success



from libcpp cimport bool as cpp_bool
from libcpp.vector cimport vector

cimport cpython.ref as cpy_ref

from CyRK.cy.common cimport CyrkErrorCodes, MAX_SIZET_SIZE
from CyRK.cy.pysolver_cyhook cimport PyEventMethod

cdef extern from "c_brentq.cpp" nogil:
    pass

cdef extern from "c_events.cpp" nogil:
    ctypedef double (*EventFunc)(double, double*, char*)

    cdef cppclass Event:
        EventFunc check
        size_t max_allowed
        size_t current_count
        size_t pyevent_index
        CyrkErrorCodes status
        double last_root
        int direction
        cpp_bool is_active
        cpp_bool initialized
        vector[double] y_at_root_vec

        Event()
        Event(
            EventFunc event_func,
            size_t max_allowed,
            int direction)

        CyrkErrorCodes setup(
            EventFunc event_func,
            size_t max_allowed,
            int direction)
        double py_check(
            double t,
            double* y_ptr,
            char* arg_ptr)
        CyrkErrorCodes set_cython_extension_instance(
            cpy_ref.PyObject* cython_extension_class_instance,
            PyEventMethod pyevent_method
            )

    ctypedef double (*EventFuncWithInst)(Event*, double, double*, char*)


# distutils: language = c++
# cython: boundscheck=False, wraparound=False, nonecheck=False, cdivision=True, initializedcheck=False

from libcpp cimport nullptr
from libcpp.vector cimport vector

from CyRK.cy.cysolver_api cimport CySolverDense
from CyRK.cy.events cimport Event
from CyRK.optimize.brentq cimport c_brentq, OptimizeInfo

# extra parameters
ctypedef struct extra_params:
    double[4] a

# callback function
cdef double f_example(Event* event_inst, double x, double* y_ptr, char *args_char_ptr) noexcept nogil:
    cdef extra_params *args_ptr = <extra_params*> args_char_ptr
    # y_ptr is unused in this example.

    # use Horner's method
    return ((args_ptr.a[3]*x + args_ptr.a[2])*x + args_ptr.a[1])*x + args_ptr.a[0]

def brentq_test(
        double a0,
        tuple args,
        double xa,
        double xb,
        double xtol,
        double rtol,
        size_t mitr):
    
    cdef vector[char] args_vec = vector[char](sizeof(extra_params))
    cdef extra_params* args_ptr = <extra_params*>args_vec.data()
    cdef OptimizeInfo optimize_info

    cdef size_t i
    for i in range(4):
        if i == 0:
            args_ptr.a[i] = a0
        else:
            args_ptr.a[i] = args[i-1]

    # Dense function is not used for this test; set to null
    cdef CySolverDense* dense_func = NULL
    cdef Event* event_inst = NULL

    # Call and return brentq root finder
    return c_brentq(f_example, xa, xb, xtol, rtol, mitr, args_vec, &optimize_info, event_inst, dense_func)



import math
import pytest

""" Same tests that SciPy uses for their optimize.brentq. """

# Solve x**3 - A0 = 0  for A0 = [2.0, 2.1, ..., 2.9].
# The ARGS have 3 elements just to show how this could be done for any cubic
# polynomial.
A0_tuple = tuple(-2.0 - x/10.0 for x in range(10))  # constant term
ARGS = (0.0, 0.0, 1.0)  # 1st, 2nd, and 3rd order terms
XLO, XHI = 0.0, 2.0  # first and second bounds of zeros functions
# absolute and relative tolerances and max iterations for zeros functions
XTOL, RTOL, MITR = 0.001, 0.001, 10
EXPECTED_list = [(-a0) ** (1.0/3.0) for a0 in A0_tuple]
# = [1.2599210498948732,
#    1.2805791649874942,
#    1.300591446851387,
#    1.3200061217959123,
#    1.338865900164339,
#    1.3572088082974532,
#    1.375068867074141,
#    1.3924766500838337,
#    1.4094597464129783,
#    1.4260431471424087]


@pytest.mark.parametrize('A0_index', range(len(A0_tuple)))
def test_brentq(A0_index):
    from CyRK.optimize.optimize_test import brentq_test

    A0 = A0_tuple[A0_index]
    expected = EXPECTED_list[A0_index]
    result = brentq_test(A0, ARGS, XLO, XHI, XTOL, RTOL, MITR)
    assert math.isclose(result, expected, rel_tol=RTOL, abs_tol=XTOL)



# distutils: language = c++
# cython: boundscheck=False, wraparound=False, nonecheck=False, cdivision=True, initializedcheck=False

# From SciPy's documentation
# Each function must return a float. The solver will find an accurate value of t at which event(t, y(t)) = 0
# using a root-finding algorithm. By default, all zeros will be found. The solver looks for a sign change over each
# step, so if multiple zero crossings occur within one step, events may be missed. "

from libc.stdio cimport printf

from libcpp cimport bool as cpp_bool
from libcpp.vector cimport vector
from libcpp.limits cimport numeric_limits

from CyRK.cy.common cimport CyrkErrorCodes, PreEvalFunc, MAX_STEP, DiffeqFuncType
from CyRK.cy.events cimport Event, EventFunc
from CyRK.cy.cysolver_api cimport cysolve_ivp_noreturn, WrapCySolverResult, ODEMethod, CySolverResult
from CyRK.cy.cysolver_test cimport lorenz_diffeq, lorenz_extraoutput_diffeq

cdef double valid_func_1(double t, double* y_ptr, char* unused_args) noexcept nogil:

    # Check if y0 is negative
    if y_ptr[0] < 0.0:
        return 0.0
    else:
        return 1.0

cdef double valid_func_2(double t, double* y_ptr, char* unused_args) noexcept nogil:

    # Check if y0 is positive
    if y_ptr[0] > 0.0:
        return 0.0
    else:
        return 1.0

cdef double valid_func_3(double t, double* y_ptr, char* unused_args) noexcept nogil:

    # Check if y1 is negative
    if y_ptr[1] < 0.0:
        return 0.0
    else:
        return 1.0


def build_event_wrapper_test():

    cdef bint tests_passed = False

    cdef EventFunc[3] event_func_array = [valid_func_1, valid_func_2, valid_func_3]
    cdef EventFunc valid_func = NULL

    cdef Event event

    cdef int build_method = -1
    cdef int[2] build_methods = [0, 1]

    cdef int direction = 0
    cdef size_t max_allowed = numeric_limits[size_t].max()

    cdef CyrkErrorCodes setup_code = CyrkErrorCodes.NO_ERROR
    cdef bint current_check = True
    cdef int build_method_tracker = 0
    cdef int func_tracker = 0
    for build_method in build_methods:
        func_tracker = 0
        for valid_func in event_func_array:                
            if build_method == 0:
                # Use constructor with all args
                event = Event(valid_func, max_allowed, direction)
            else:
                # Build empty event and use setup
                event = Event()
                setup_code = event.setup(valid_func, max_allowed, direction)
                current_check = current_check and (setup_code == CyrkErrorCodes.NO_ERROR)
            current_check = current_check and event.initialized and (event.status == CyrkErrorCodes.NO_ERROR)
            if not current_check:
                printf("TEST FAILED: `build_event_wrapper_test` - building events failed at build method %d and event func %d.", build_method_tracker, func_tracker)
                break
            func_tracker += 1
        build_method_tracker += 1

    tests_passed = current_check

    return tests_passed


cdef double lorenz_event_func_1(double t, double* y_ptr, char* unused_args) noexcept nogil:

    # Check if t greater than or equal to 5.0
    if t >= 5.0:
        return 0.0
    else:
        return 1.0

cdef double lorenz_event_func_2(double t, double* y_ptr, char* unused_args) noexcept nogil:

    # Check y values.
    # In the time span [0,10]: 
    #    y_0  starts at 1, spikes then goes below zero and oscillates with a min below -10. Have this return if y_0 < -10
    if y_ptr[0] < -10.0:
        return 0.0
    elif y_ptr[2] > 30.0:
        return 0.0
    else:
        return 1.0

cdef double lorenz_event_func_3(double t, double* y_ptr, char* args) noexcept nogil:

    # Use arguments to set threshold
    cdef double* args_dbl_ptr = <double*>args

    # We won't actually use the args but lets just check they are correct.
    cdef cpp_bool args_correct = False
    if args_dbl_ptr[0] == 10.0 and args_dbl_ptr[1] == 28.0 and args_dbl_ptr[2] == 8.0 / 3.0:
        args_correct = True

    # Then return events if args are correct and t greater than 8
    if args_correct:
        return 0.0
    else:
        return 1.0


def run_cysolver_with_events(
        bint use_dense,
        double[::1] t_eval,
        bint use_termination,
        bint capture_extra
    ):

    cdef bint tests_passed = True

    # Inputs for lorenz diffeq
    cdef size_t num_y = 3
    cdef vector[double] y0_vec = vector[double](num_y)
    y0_vec[0] = 1.0
    y0_vec[1] = 0.0
    y0_vec[2] = 0.0

    cdef vector[double] t_eval_vec = vector[double]()
    cdef cpp_bool t_eval_provided = False
    cdef size_t i
    if t_eval.size > 0:
        t_eval_provided = True
        for i in range(t_eval.size):
            t_eval_vec.push_back(t_eval[i])

    cdef double t_start = 0.0
    cdef double t_end = 10.0

    cdef vector[char] args_vec = vector[char](3 * sizeof(double))
    args_dbl_ptr = <double*>args_vec.data()
    args_dbl_ptr[0] = 10.0
    args_dbl_ptr[1] = 28.0
    args_dbl_ptr[2] = 8.0 / 3.0

    cdef size_t num_dy = 3

    cdef PreEvalFunc pre_eval_func = NULL
    
    # Build events
    cdef vector[Event] events_vec = vector[Event]()
    cdef size_t max_allowed = numeric_limits[size_t].max()
    cdef int direction = 0
    if use_termination:
        events_vec.emplace_back(lorenz_event_func_1, 1, 0)
    else:
        events_vec.emplace_back(lorenz_event_func_1, max_allowed, direction)
    events_vec.emplace_back(lorenz_event_func_2, max_allowed, direction)
    events_vec.emplace_back(lorenz_event_func_3, max_allowed, direction)

    cdef ODEMethod integration_method = ODEMethod.RK45
    cdef WrapCySolverResult solution = WrapCySolverResult()
    solution.build_cyresult(integration_method)
    cdef CySolverResult* solution_ptr = solution.cyresult_uptr.get()

    cdef DiffeqFuncType diffeq = lorenz_diffeq
    cdef size_t num_extra = 0
    if capture_extra:
        diffeq = lorenz_extraoutput_diffeq
        num_extra = 3
        num_dy = 6

    # Run Solver
    cysolve_ivp_noreturn(
        solution_ptr,
        diffeq,
        t_start,
        t_end,
        y0_vec,
        rtol = 1.0e-3,
        atol = 1.0e-6,
        args_vec = args_vec,
        num_extra = num_extra,
        max_num_steps = 100000,
        max_ram_MB = 2000,
        dense_output = use_dense,
        t_eval_vec = t_eval_vec,
        pre_eval_func = pre_eval_func,
        events_vec = events_vec,
        rtols_vec = vector[double](),
        atols_vec = vector[double](),
        max_step = MAX_STEP,
        first_step = 0.0,
        expected_size = 128
        )
    
    solution.finalize()
    if not solution_ptr.success:
        printf("TEST FAILED: `run_cysolver_with_events` solver did not complete successfully.\n")
        tests_passed = False
    
    # Check that solution storage is properly setup to handle events
    if solution_ptr.num_events != 3:
        printf("TEST FAILED: `run_cysolver_with_events` - solution_ptr.num_events != 3 (%zd).\n", solution_ptr.num_events)
        tests_passed = False
    if solution_ptr.event_times.size() != 3:
        printf("TEST FAILED: `run_cysolver_with_events` - solution_ptr.event_times.size() != 3 (%zd).\n", solution_ptr.event_times.size())
        tests_passed = False
    if solution_ptr.event_states.size() != 3:
        printf("TEST FAILED: `run_cysolver_with_events` - solution_ptr.event_states.size() != 3 (%zd).\n", solution_ptr.event_states.size())
        tests_passed = False

    # Check Solution event data
    if solution_ptr.num_events != 3:
        printf("TEST FAILED: `run_cysolver_with_events` - solution_ptr.num_events != 3 (%zd).\n", solution_ptr.num_events)
        tests_passed = False
    
    # Check CySolver for correct event data
    if solution_ptr.solver_uptr.get().num_events != 3:
        printf("TEST FAILED: `run_cysolver_with_events` - solution_ptr.solver_uptr.get().num_events != 3 (%zd).\n", solution_ptr.solver_uptr.get().num_events)
        tests_passed = False
    
    # Check results of event run.
    if solution_ptr.event_times.size() != 3:
        printf("TEST FAILED: `run_cysolver_with_events` - solution_ptr.event_times.size() != 3 (%zd).\n", solution_ptr.event_times.size())
        tests_passed = False
    if solution_ptr.event_states.size() != 3:
        printf("TEST FAILED: `run_cysolver_with_events` - solution_ptr.event_states.size() != 3 (%zd).\n", solution_ptr.event_states.size())
        tests_passed = False
    
    cdef size_t event_i
    for event_i in range(3):
        printf("INFO: `run_cysolver_with_events` - solution_ptr.event_times[%zd].size() = %zd\n", event_i, solution_ptr.event_times[event_i].size())
        if solution_ptr.event_times[event_i].size() == 0:
            printf("TEST FAILED: `run_cysolver_with_events` - solution_ptr.event_times[%zd].size() == 0.\n", event_i)
            tests_passed = False
            break
        
        printf("INFO: `run_cysolver_with_events` - solution_ptr.event_states[%zd].size() = %d\n", event_i, solution_ptr.event_states[event_i].size())
        if solution_ptr.event_states[event_i].size() == 0:
            printf("TEST FAILED: `run_cysolver_with_events` - solution_ptr.event_states[%zd].size() == 0.\n", event_i)
            tests_passed = False
            break
    
        if solution_ptr.event_states[event_i].size() != num_dy * solution_ptr.event_times[event_i].size():
            printf("TEST FAILED: `run_cysolver_with_events` - solution_ptr.event_states[%zd].size() / num_dy (%zd) != event_times.\n", event_i, num_dy)
            tests_passed = False
            break

    cdef double time_check = 0.0
    if use_termination:
        if not solution_ptr.event_terminated:
            printf("TEST FAILED: `run_cysolver_with_events` - solution.event_terminated is False.\n")
            tests_passed = False

        time_check = solution_ptr.time_domain_vec.back()

        # Surprisingly scipy's root finding method really does not get very close to 5.0 so we will check if its above 5.1; for this example it should be ~5.035
        if time_check > 5.1:
            printf("TEST FAILED: `run_cysolver_with_events` - Solver did not terminated when requested: t_end = %f.\n", time_check)
            tests_passed = False
    
    if not tests_passed:
        printf("`run_cysolver_with_events` - Tests Failed. use_dense = %d; t_eval_provided = %d; use_termination = %d\n", use_dense, t_eval_provided, use_termination)
    
    return solution, tests_passed



